from __future__ import (division)

from pomegranate import *
from pomegranate.io import DataGenerator
from pomegranate.io import DataFrameGenerator

from nose.tools import with_setup
from nose.tools import assert_almost_equal
from nose.tools import assert_equal
from nose.tools import assert_not_equal
from nose.tools import assert_less_equal
from nose.tools import assert_raises
from nose.tools import assert_true
from numpy.testing import assert_array_equal
from numpy.testing import assert_array_almost_equal

import pandas
import random
import pickle
import numpy as np

nan = numpy.nan

def setup_univariate_mixed():
	normal = NormalDistribution(5, 2)
	uniform = UniformDistribution(0, 10)

	global model
	model = NaiveBayes([normal, uniform])

	global X
	X = numpy.array([[5], [3], [1], [-1]])

def setup_multivariate_gaussian():
	d11 = NormalDistribution(0.0, 1)
	d12 = NormalDistribution(0.5, 1)
	d13 = NormalDistribution(0.3, 1)
	d1 = IndependentComponentsDistribution([d11, d12, d13])

	d21 = NormalDistribution(1.0, 1)
	d22 = NormalDistribution(1.2, 1)
	d23 = NormalDistribution(1.5, 1)
	d2 = IndependentComponentsDistribution([d21, d22, d23])

	global model
	model = NaiveBayes([d1, d2])

	global X
	X = numpy.array([[0.3, 0.5, 0.1],
					 [0.8, 1.4, 0.5],
					 [1.4, 2.6, 1.8],
					 [4.2, 3.3, 3.7],
					 [2.6, 3.6, 3.3]])

	global y
	y = [0, 0, 0, 1, 1]

	global X_nan
	X_nan = numpy.array([[0.3, nan, 0.1],
		     			 [nan, 1.4, nan],
			     		 [1.4, 2.6, nan],
				    	 [nan, nan, nan],
					     [nan, 3.6, 3.3]])


def setup_multivariate_mixed():
	d11 = ExponentialDistribution(5)
	d12 = LogNormalDistribution(0.5, 0.78)
	d13 = PoissonDistribution(4)
	d1 = IndependentComponentsDistribution([d11, d12, d13])

	d21 = ExponentialDistribution(35)
	d22 = LogNormalDistribution(1.8, 1.33)
	d23 = PoissonDistribution(6)
	d2 = IndependentComponentsDistribution([d21, d22, d23])

	global model
	model = NaiveBayes([d1, d2])

	global X
	X = numpy.array([[0.3, 0.5, 0.1],
					 [0.8, 1.4, 0.5],
					 [1.4, 2.6, 1.8],
					 [4.2, 3.3, 3.7],
					 [2.6, 3.6, 3.3]])

	global y
	y = [0, 0, 0, 1, 1]

	global X_nan
	X_nan = numpy.array([[0.3, nan, 0.1],
		     			 [nan, 1.4, nan],
			     		 [1.4, 2.6, nan],
				    	 [nan, nan, nan],
					     [nan, 3.6, 3.3]])


def teardown():
	pass


@with_setup(setup_univariate_mixed, teardown)
def test_nb_univariate_initialization():
	assert_equal(model.d, 1)
	assert_equal(model.n, 2)


@with_setup(setup_multivariate_mixed, teardown)
def test_nb_multivariate_initialization():
	assert_equal(model.d, 3)
	assert_equal(model.n, 2)


def test_nb_univariate_constructors():
	d1 = NormalDistribution(0.5, 1)
	d2 = MultivariateGaussianDistribution([0, 0], [[1, 0], [0, 1]])
	d3 = IndependentComponentsDistribution([NormalDistribution(0, 1),
		NormalDistribution(2, 1), NormalDistribution(3, 1)])

	assert_raises(TypeError, NaiveBayes, [d1, d2])
	assert_raises(TypeError, NaiveBayes, [d1, d3])
	assert_raises(ValueError, NaiveBayes, [NormalDistribution])


def test_nb_multivariate_constructors():
	d1 = MultivariateGaussianDistribution([0, 0], [[1, 0], [0, 1]])
	d2 = IndependentComponentsDistribution([NormalDistribution(0, 1),
		NormalDistribution(2, 1), NormalDistribution(3, 1)])
	d3 = IndependentComponentsDistribution([NormalDistribution(0, 1),
		NormalDistribution(2, 1)])

	NaiveBayes([d1, d3])
	assert_raises(TypeError, NaiveBayes, [d2, d3])
	assert_raises(TypeError, NaiveBayes, [d2, d1])
	assert_raises(ValueError, NaiveBayes, [MultivariateGaussianDistribution])
	assert_raises(ValueError, NaiveBayes, [IndependentComponentsDistribution])


@with_setup(setup_univariate_mixed, teardown)
def test_nb_univariate_mixed_predict_log_proba():
	y_hat = model.predict_log_proba(X)
	y = [[-0.4063484, -1.096847],
	     [-0.6024268, -0.792926],
   	  	 [-1.5484819, -0.238981],
		 [ 0.0,   float('-inf')]]

	assert_array_almost_equal(y, y_hat)


@with_setup(setup_multivariate_gaussian, teardown)
def test_nb_multivariate_gaussian_predict_log_proba():
	y_hat = model.predict_log_proba(X)
	y = [[ -2.194303e-01,    -1.624430e+00],
 		 [ -8.00891133e-01,  -5.95891133e-01],
		 [ -3.24475797e+00,  -3.97579742e-02],
 		 [ -8.77515454e+00,  -1.54536960e-04],
 		 [ -6.90600226e+00,  -1.00225665e-03]]

	assert_array_almost_equal(y, y_hat)


@with_setup(setup_multivariate_mixed, teardown)
def test_nb_multivariate_mixed_predict_log_proba():
	y_hat = model.predict_log_proba(X)
	y = [[ -3.96979060e-05,  -1.01342320e+01],
		 [ -1.43325352e-11,  -2.49684574e+01],
		 [  0.00000000e+00,  -4.18889545e+01],
		 [  0.00000000e+00,  -1.24795606e+02],
		 [  0.00000000e+00,  -7.68246547e+01]]

	assert_array_almost_equal(y, y_hat)


@with_setup(setup_multivariate_gaussian, teardown)
def test_nb_multivariate_gaussian_nan_predict_log_proba():
	y_hat = model.predict_log_proba(X_nan)
	y = [[-0.27268481, -1.43268481],
		 [-0.90406199, -0.51906199],
		 [-2.23782228, -0.11282228],
		 [-0.69314718, -0.69314718],
		 [-4.81315536, -0.00815536]]

	assert_array_almost_equal(y, y_hat)


@with_setup(setup_multivariate_mixed, teardown)
def test_nb_multivariate_mixed_nan_predict_log_proba():
	y_hat = model.predict_log_proba(X_nan)
	y = [[ -1.21742279e-04,  -9.01366508e+00],
		 [ -2.83092062e-01,  -1.40019217e+00],
		 [  0.00000000e+00,  -4.06187917e+01],
		 [ -6.93147181e-01,  -6.93147181e-01],
		 [ -3.80319311e-01,  -1.15088421e+00]]

	assert_array_almost_equal(y, y_hat)


@with_setup(setup_univariate_mixed, teardown)
def test_nb_univariate_mixed_predict_log_proba_parallel():
	y_hat = model.predict_log_proba(X, n_jobs=2)
	y = [[-0.4063484, -1.096847],
	     [-0.6024268, -0.792926],
   	  	 [-1.5484819, -0.238981],
		 [ 0.0,   float('-inf')]]

	assert_array_almost_equal(y, y_hat)


@with_setup(setup_multivariate_gaussian, teardown)
def test_nb_multivariate_gaussian_predict_log_proba_parallel():
	y_hat = model.predict_log_proba(X, n_jobs=2)
	y = [[ -2.194303e-01,    -1.624430e+00],
 		 [ -8.00891133e-01,  -5.95891133e-01],
		 [ -3.24475797e+00,  -3.97579742e-02],
 		 [ -8.77515454e+00,  -1.54536960e-04],
 		 [ -6.90600226e+00,  -1.00225665e-03]]

	assert_array_almost_equal(y, y_hat)


@with_setup(setup_multivariate_mixed, teardown)
def test_nb_multivariate_mixed_predict_log_proba_parallel():
	y_hat = model.predict_log_proba(X, n_jobs=2)
	y = [[ -3.96979060e-05,  -1.01342320e+01],
		 [ -1.43325352e-11,  -2.49684574e+01],
		 [  0.00000000e+00,  -4.18889545e+01],
		 [  0.00000000e+00,  -1.24795606e+02],
		 [  0.00000000e+00,  -7.68246547e+01]]

	assert_array_almost_equal(y, y_hat)


@with_setup(setup_univariate_mixed, teardown)
def test_nb_univariate_mixed_predict_proba():
	y_hat = model.predict_proba(X)
	y = [[ 0.66607801,  0.33392199],
		 [ 0.54748134,  0.45251866],
		 [ 0.21257042,  0.78742958],
		 [ 1.,          0.        ]]

	assert_array_almost_equal(y, y_hat)


@with_setup(setup_multivariate_gaussian, teardown)
def test_nb_multivariate_gaussian_predict_proba():
	y_hat = model.predict_proba(X)
	y =	[[  8.02976114e-01,   1.97023886e-01],
		 [  4.48928731e-01,   5.51071269e-01],
		 [  3.89779969e-02,   9.61022003e-01],
		 [  1.54525019e-04,   9.99845475e-01],
		 [  1.00175456e-03,   9.98998245e-01]]

	assert_array_almost_equal(y, y_hat)


@with_setup(setup_multivariate_mixed, teardown)
def test_nb_multivariate_mixed_predict_proba():
	y_hat = model.predict_proba(X)
	y = [[  9.99960303e-01,   3.96971181e-05],
		 [  1.00000000e+00,   1.43329876e-11],
		 [  1.00000000e+00,   6.42477904e-19],
		 [  1.00000000e+00,   6.33806932e-55],
		 [  1.00000000e+00,   4.31992661e-34]]

	assert_array_almost_equal(y, y_hat)


@with_setup(setup_multivariate_gaussian, teardown)
def test_nb_multivariate_gaussian_nan_predict_proba():
	y_hat = model.predict_proba(X_nan)
	y = [[ 0.76133271,  0.23866729],
		 [ 0.40492153,  0.59507847],
		 [ 0.10669059,  0.89330941],
		 [ 0.5,         0.5       ],
		 [ 0.00812219,  0.99187781]]

	assert_array_almost_equal(y, y_hat)


@with_setup(setup_multivariate_mixed, teardown)
def test_nb_multivariate_mixed_nan_predict_proba():
	y_hat = model.predict_proba(X_nan)
	y = [[  9.99878265e-01,   1.21734869e-04],
		 [  7.53450421e-01,   2.46549579e-01],
		 [  1.00000000e+00,   2.28814158e-18],
		 [  5.00000000e-01,   5.00000000e-01],
		 [  6.83643080e-01,   3.16356920e-01]]

	assert_array_almost_equal(y, y_hat)


@with_setup(setup_univariate_mixed, teardown)
def test_nb_univariate_mixed_predict_proba_parallel():
	y_hat = model.predict_proba(X, n_jobs=2)
	y =  [[ 0.66607801,  0.33392199],
		  [ 0.54748134,  0.45251866],
		  [ 0.21257042,  0.78742958],
		  [ 1.,          0.        ]]

	assert_array_almost_equal(y, y_hat)


@with_setup(setup_multivariate_gaussian, teardown)
def test_nb_multivariate_gaussian_predict_proba_parallel():
	y_hat = model.predict_proba(X, n_jobs=2)
	y = [[  8.02976114e-01,   1.97023886e-01],
		 [  4.48928731e-01,   5.51071269e-01],
		 [  3.89779969e-02,   9.61022003e-01],
		 [  1.54525019e-04,   9.99845475e-01],
		 [  1.00175456e-03,   9.98998245e-01]]

	assert_array_almost_equal(y, y_hat)


@with_setup(setup_multivariate_mixed, teardown)
def test_nb_multivariate_mixed_predict_proba_parallel():
	y_hat = model.predict_proba(X, n_jobs=2)
	y = [[  9.99960303e-01,   3.96971181e-05,],
		 [  1.00000000e+00,   1.43329876e-11,],
		 [  1.00000000e+00,   6.42477904e-19,],
		 [  1.00000000e+00,   6.33806932e-55,],
		 [  1.00000000e+00,   4.31992661e-34,]]

	assert_array_almost_equal(y, y_hat)


@with_setup(setup_univariate_mixed, teardown)
def test_nb_univariate_mixed_predict():
	y_hat = model.predict(X)
	y = [0, 0, 1, 0]

	assert_array_almost_equal(y, y_hat)


@with_setup(setup_multivariate_gaussian, teardown)
def test_nb_multivariate_gaussian_predict():
	y_hat = model.predict(X)
	y = [0, 1, 1, 1, 1]

	assert_array_almost_equal(y, y_hat)


@with_setup(setup_multivariate_mixed, teardown)
def test_nb_multivariate_mixed_predict():
	y_hat = model.predict(X)
	y = [0, 0, 0, 0, 0]

	assert_array_almost_equal(y, y_hat)


@with_setup(setup_multivariate_gaussian, teardown)
def test_nb_multivariate_gaussian_nan_predict():
	y_hat = model.predict(X_nan)
	y = [0, 1, 1, 0, 1]

	assert_array_almost_equal(y, y_hat)


@with_setup(setup_multivariate_mixed, teardown)
def test_nb_multivariate_mixed_nan_predict():
	y_hat = model.predict(X_nan)
	y = [0, 0, 0, 0, 0]

	assert_array_almost_equal(y, y_hat)


@with_setup(setup_univariate_mixed, teardown)
def test_nb_univariate_mixed_predict_parallel():
	y_hat = model.predict(X, n_jobs=2)
	y = [0, 0, 1, 0]

	assert_array_almost_equal(y, y_hat)


@with_setup(setup_multivariate_gaussian, teardown)
def test_nb_multivariate_gaussian_predict_parallel():
	y_hat = model.predict(X, n_jobs=2)
	y = [0, 1, 1, 1, 1]

	assert_array_almost_equal(y, y_hat)


@with_setup(setup_multivariate_mixed, teardown)
def test_nb_multivariate_mixed_predict_parallel():
	y_hat = model.predict(X, n_jobs=2)
	y = [0, 0, 0, 0, 0]

	assert_array_almost_equal(y, y_hat)


@with_setup(setup_univariate_mixed, teardown)
def test_nb_univariate_mixed_fit():
	X = np.array([5, 4, 5, 4, 6, 5, 6, 5, 4, 6, 5, 4, 0, 0,
		1, 9, 8, 2, 0, 1, 1, 8, 10, 0]).reshape(-1, 1)
	y = np.array([0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 1,
		1, 1, 1, 1, 1, 1, 1, 1, 1, 1])

	model.fit(X, y)

	d1 = model.distributions[0]
	d2 = model.distributions[1]

	assert_array_almost_equal(d1.parameters, [4.916666666666667, 0.7592027982620252])
	assert_array_almost_equal(d2.parameters, [0.0, 10.0])

@with_setup(setup_multivariate_gaussian, teardown)
def test_nb_multivariate_gaussian_fit():
	model.fit(X, y)

	d11 = model.distributions[0].distributions[0]
	d12 = model.distributions[0].distributions[1]
	d13 = model.distributions[0].distributions[2]
	d21 = model.distributions[1].distributions[0]
	d22 = model.distributions[1].distributions[1]
	d23 = model.distributions[1].distributions[2]

	assert_array_almost_equal(d11.parameters, [0.8333333333333334, 0.4496912521077347])
	assert_array_almost_equal(d12.parameters, [1.5, 0.8602325267042628])
	assert_array_almost_equal(d13.parameters, [0.7999999999999999, 0.725718035235908])
	assert_array_almost_equal(d21.parameters, [3.4000000000000004, 0.7999999999999993])
	assert_array_almost_equal(d22.parameters, [3.45, 0.1499999999999969])
	assert_array_almost_equal(d23.parameters, [3.5, 0.19999999999999787])


@with_setup(setup_multivariate_mixed, teardown)
def test_nb_multivariate_mixed_fit():
	model.fit(X, y)

	d11 = model.distributions[0].distributions[0]
	d12 = model.distributions[0].distributions[1]
	d13 = model.distributions[0].distributions[2]
	d21 = model.distributions[1].distributions[0]
	d22 = model.distributions[1].distributions[1]
	d23 = model.distributions[1].distributions[2]

	assert_array_almost_equal(d11.parameters, [1.1999999920000004])
	assert_array_almost_equal(d12.parameters, [0.199612167029568, 0.6799837375101412])
	assert_array_almost_equal(d13.parameters, [0.7999999999999999])
	assert_array_almost_equal(d21.parameters, [0.2941176574394461])
	assert_array_almost_equal(d22.parameters, [1.2374281569672494, 0.04350568849481522])
	assert_array_almost_equal(d23.parameters, [3.5])


@with_setup(setup_multivariate_gaussian, teardown)
def test_nb_multivariate_gaussian_nan_fit():
	model.fit(X_nan, y)

	d11 = model.distributions[0].distributions[0]
	d12 = model.distributions[0].distributions[1]
	d13 = model.distributions[0].distributions[2]
	d21 = model.distributions[1].distributions[0]
	d22 = model.distributions[1].distributions[1]
	d23 = model.distributions[1].distributions[2]

	assert_array_almost_equal(d11.parameters, [0.85, 0.55])
	assert_array_almost_equal(d12.parameters, [2.0, 0.6000000000000003])
	assert_array_almost_equal(d13.parameters, [0.1, 0.0])
	assert_array_almost_equal(d21.parameters, [1.0, 1.0])
	assert_array_almost_equal(d22.parameters, [3.6, 0.0])
	assert_array_almost_equal(d23.parameters, [3.3, 0.0])


@with_setup(setup_multivariate_mixed, teardown)
def test_nb_multivariate_mixed_nan_fit():
	model.fit(X_nan, y)

	d11 = model.distributions[0].distributions[0]
	d12 = model.distributions[0].distributions[1]
	d13 = model.distributions[0].distributions[2]
	d21 = model.distributions[1].distributions[0]
	d22 = model.distributions[1].distributions[1]
	d23 = model.distributions[1].distributions[2]

	assert_array_almost_equal(d11.parameters, [1.1764705778546718])
	assert_array_almost_equal(d12.parameters, [0.645991, 0.3095196])
	assert_array_almost_equal(d13.parameters, [0.1])
	assert_array_almost_equal(d21.parameters, [35.0])
	assert_array_almost_equal(d22.parameters, [1.2809338454620642, 0.0])
	assert_array_almost_equal(d23.parameters, [3.3])


@with_setup(setup_univariate_mixed, teardown)
def test_nb_univariate_mixed_fit_parallel():
	X = np.array([5, 4, 5, 4, 6, 5, 6, 5, 4, 6, 5, 4, 0, 0,
		1, 9, 8, 2, 0, 1, 1, 8, 10, 0]).reshape(-1, 1)
	y = np.array([0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 1,
		1, 1, 1, 1, 1, 1, 1, 1, 1, 1])

	model.fit(X, y, n_jobs=2)

	d1 = model.distributions[0]
	d2 = model.distributions[1]

	assert_array_almost_equal(d1.parameters, [4.916666666666667, 0.7592027982620252])
	assert_array_almost_equal(d2.parameters, [0.0, 10.0])

@with_setup(setup_multivariate_gaussian, teardown)
def test_nb_multivariate_gaussian_fit_parallel():
	model.fit(X, y, n_jobs=2)

	d11 = model.distributions[0].distributions[0]
	d12 = model.distributions[0].distributions[1]
	d13 = model.distributions[0].distributions[2]
	d21 = model.distributions[1].distributions[0]
	d22 = model.distributions[1].distributions[1]
	d23 = model.distributions[1].distributions[2]

	assert_array_almost_equal(d11.parameters, [0.8333333333333334, 0.4496912521077347])
	assert_array_almost_equal(d12.parameters, [1.5, 0.8602325267042628])
	assert_array_almost_equal(d13.parameters, [0.7999999999999999, 0.725718035235908])
	assert_array_almost_equal(d21.parameters, [3.4000000000000004, 0.7999999999999993])
	assert_array_almost_equal(d22.parameters, [3.45, 0.1499999999999969])
	assert_array_almost_equal(d23.parameters, [3.5, 0.19999999999999787])


@with_setup(setup_multivariate_mixed, teardown)
def test_nb_multivariate_mixed_fit_parallel():
	model.fit(X, y, n_jobs=2)

	d11 = model.distributions[0].distributions[0]
	d12 = model.distributions[0].distributions[1]
	d13 = model.distributions[0].distributions[2]
	d21 = model.distributions[1].distributions[0]
	d22 = model.distributions[1].distributions[1]
	d23 = model.distributions[1].distributions[2]

	assert_array_almost_equal(d11.parameters, [1.1999999920000004])
	assert_array_almost_equal(d12.parameters, [0.199612167029568, 0.6799837375101412])
	assert_array_almost_equal(d13.parameters, [0.7999999999999999])
	assert_array_almost_equal(d21.parameters, [0.2941176574394461])
	assert_array_almost_equal(d22.parameters, [1.2374281569672494, 0.04350568849481522])
	assert_array_almost_equal(d23.parameters, [3.5])


@with_setup(setup_univariate_mixed, teardown)
def test_nb_univariate_mixed_from_samples():
	X = np.array([5, 4, 5, 4, 6, 5, 6, 5, 4, 6, 5, 4, 0, 0,
		1, 9, 8, 2, 0, 1, 1, 8, 10, 0]).reshape(-1, 1)
	y = np.array([0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 1,
		1, 1, 1, 1, 1, 1, 1, 1, 1, 1])

	model = NaiveBayes.from_samples([NormalDistribution, UniformDistribution],
		X, y)

	d1 = model.distributions[0]
	d2 = model.distributions[1]

	assert_array_almost_equal(d1.parameters, [4.916666666666667, 0.7592027982620252])
	assert_array_almost_equal(d2.parameters, [0.0, 10.0])

@with_setup(setup_multivariate_gaussian, teardown)
def test_nb_multivariate_gaussian_from_samples():
	model = NaiveBayes.from_samples(NormalDistribution, X, y)

	d11 = model.distributions[0].distributions[0]
	d12 = model.distributions[0].distributions[1]
	d13 = model.distributions[0].distributions[2]
	d21 = model.distributions[1].distributions[0]
	d22 = model.distributions[1].distributions[1]
	d23 = model.distributions[1].distributions[2]

	assert_array_almost_equal(d11.parameters, [0.8333333333333334, 0.4496912521077347])
	assert_array_almost_equal(d12.parameters, [1.5, 0.8602325267042628])
	assert_array_almost_equal(d13.parameters, [0.7999999999999999, 0.725718035235908])
	assert_array_almost_equal(d21.parameters, [3.4000000000000004, 0.7999999999999993])
	assert_array_almost_equal(d22.parameters, [3.45, 0.1499999999999969])
	assert_array_almost_equal(d23.parameters, [3.5, 0.19999999999999787])


@with_setup(setup_multivariate_mixed, teardown)
def test_nb_multivariate_mixed_from_samples():
	d = [ExponentialDistribution, LogNormalDistribution, PoissonDistribution]
	model = NaiveBayes.from_samples(d, X, y)

	d11 = model.distributions[0].distributions[0]
	d12 = model.distributions[0].distributions[1]
	d13 = model.distributions[0].distributions[2]
	d21 = model.distributions[1].distributions[0]
	d22 = model.distributions[1].distributions[1]
	d23 = model.distributions[1].distributions[2]

	assert_array_almost_equal(d11.parameters, [1.1999999920000004])
	assert_array_almost_equal(d12.parameters, [0.199612167029568, 0.6799837375101412])
	assert_array_almost_equal(d13.parameters, [0.7999999999999999])
	assert_array_almost_equal(d21.parameters, [0.2941176574394461])
	assert_array_almost_equal(d22.parameters, [1.2374281569672494, 0.04350568849481522])
	assert_array_almost_equal(d23.parameters, [3.5])


@with_setup(setup_multivariate_gaussian, teardown)
def test_nb_multivariate_gaussian_nan_from_samples():
	model = NaiveBayes.from_samples(NormalDistribution, X_nan, y)

	d11 = model.distributions[0].distributions[0]
	d12 = model.distributions[0].distributions[1]
	d13 = model.distributions[0].distributions[2]
	d21 = model.distributions[1].distributions[0]
	d22 = model.distributions[1].distributions[1]
	d23 = model.distributions[1].distributions[2]

	assert_array_almost_equal(d11.parameters, [0.85, 0.55])
	assert_array_almost_equal(d12.parameters, [2.0, 0.6000000000000003])
	assert_array_almost_equal(d13.parameters, [0.1, 0.0])
	assert_array_almost_equal(d21.parameters, [0.0, 1.0])
	assert_array_almost_equal(d22.parameters, [3.6, 0.0])
	assert_array_almost_equal(d23.parameters, [3.3, 0.0])


@with_setup(setup_multivariate_mixed, teardown)
def test_nb_multivariate_mixed_nan_from_samples():
	d = [ExponentialDistribution, LogNormalDistribution, PoissonDistribution]
	model = NaiveBayes.from_samples(d, X_nan, y)

	d11 = model.distributions[0].distributions[0]
	d12 = model.distributions[0].distributions[1]
	d13 = model.distributions[0].distributions[2]
	d21 = model.distributions[1].distributions[0]
	d22 = model.distributions[1].distributions[1]
	d23 = model.distributions[1].distributions[2]

	assert_array_almost_equal(d11.parameters, [1.1764705778546718])
	assert_array_almost_equal(d12.parameters, [0.645991, 0.3095196])
	assert_array_almost_equal(d13.parameters, [0.1])
	assert_array_almost_equal(d21.parameters, [1])
	assert_array_almost_equal(d22.parameters, [1.2809338454620642, 0.0])
	assert_array_almost_equal(d23.parameters, [3.3])


@with_setup(setup_univariate_mixed, teardown)
def test_nb_univariate_mixed_from_samples_parallel():
	X = np.array([5, 4, 5, 4, 6, 5, 6, 5, 4, 6, 5, 4, 0, 0,
		1, 9, 8, 2, 0, 1, 1, 8, 10, 0]).reshape(-1, 1)
	y = np.array([0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 1,
		1, 1, 1, 1, 1, 1, 1, 1, 1, 1])

	model = NaiveBayes.from_samples([NormalDistribution, UniformDistribution],
		X, y, n_jobs=2)

	d1 = model.distributions[0]
	d2 = model.distributions[1]

	assert_array_almost_equal(d1.parameters, [4.916666666666667, 0.7592027982620252])
	assert_array_almost_equal(d2.parameters, [0.0, 10.0])

@with_setup(setup_multivariate_gaussian, teardown)
def test_nb_multivariate_gaussian_from_samples_parallel():
	model = NaiveBayes.from_samples(NormalDistribution, X, y, n_jobs=2)

	d11 = model.distributions[0].distributions[0]
	d12 = model.distributions[0].distributions[1]
	d13 = model.distributions[0].distributions[2]
	d21 = model.distributions[1].distributions[0]
	d22 = model.distributions[1].distributions[1]
	d23 = model.distributions[1].distributions[2]

	assert_array_almost_equal(d11.parameters, [0.8333333333333334, 0.4496912521077347])
	assert_array_almost_equal(d12.parameters, [1.5, 0.8602325267042628])
	assert_array_almost_equal(d13.parameters, [0.7999999999999999, 0.725718035235908])
	assert_array_almost_equal(d21.parameters, [3.4000000000000004, 0.7999999999999993])
	assert_array_almost_equal(d22.parameters, [3.45, 0.1499999999999969])
	assert_array_almost_equal(d23.parameters, [3.5, 0.19999999999999787])


@with_setup(setup_multivariate_mixed, teardown)
def test_nb_multivariate_mixed_from_samples_parallel():
	d = [ExponentialDistribution, LogNormalDistribution, PoissonDistribution]
	model = NaiveBayes.from_samples(d, X, y, n_jobs=2)

	d11 = model.distributions[0].distributions[0]
	d12 = model.distributions[0].distributions[1]
	d13 = model.distributions[0].distributions[2]
	d21 = model.distributions[1].distributions[0]
	d22 = model.distributions[1].distributions[1]
	d23 = model.distributions[1].distributions[2]

	assert_array_almost_equal(d11.parameters, [1.1999999920000004])
	assert_array_almost_equal(d12.parameters, [0.199612167029568, 0.6799837375101412])
	assert_array_almost_equal(d13.parameters, [0.7999999999999999])
	assert_array_almost_equal(d21.parameters, [0.2941176574394461])
	assert_array_almost_equal(d22.parameters, [1.2374281569672494, 0.04350568849481522])
	assert_array_almost_equal(d23.parameters, [3.5])


@with_setup(setup_univariate_mixed, teardown)
def test_raise_errors():
	# check raises no errors when converting values
	model.predict_log_proba([[5]])
	model.predict_log_proba([[4.5]])
	model.predict_log_proba([[5], [6]])
	model.predict_log_proba(np.array([[5], [6]]) )

	model.predict_proba([[5]])
	model.predict_proba([[4.5]])
	model.predict_proba([[5], [6]])
	model.predict_proba(np.array([[5], [6]]))

	model.predict([[5]])
	model.predict([[4.5]])
	model.predict([[5], [6]])
	model.predict(np.array([[5], [6]]))

@with_setup(setup_univariate_mixed, teardown)
def test_pickling():
	j_univ = pickle.dumps(model)

	new_univ = pickle.loads(j_univ)
	assert_true(isinstance(new_univ.distributions[0], NormalDistribution))
	assert_true(isinstance(new_univ.distributions[1], UniformDistribution))
	assert_true(isinstance(new_univ, NaiveBayes))
	numpy.testing.assert_array_equal(model.weights, new_univ.weights)

@with_setup(setup_univariate_mixed, teardown)
def test_json():
	j_univ = model.to_json()

	new_univ = model.from_json(j_univ)
	assert_true(isinstance(new_univ.distributions[0], NormalDistribution))
	assert_true(isinstance(new_univ.distributions[1], UniformDistribution))
	assert_true(isinstance(new_univ, NaiveBayes))
	numpy.testing.assert_array_equal( model.weights, new_univ.weights)

@with_setup(setup_univariate_mixed, teardown)
def test_robust_from_json():
	j_univ = model.to_json()

	new_univ = from_json(j_univ)
	assert_true(isinstance(new_univ.distributions[0], NormalDistribution))
	assert_true(isinstance(new_univ.distributions[1], UniformDistribution))
	assert_true(isinstance(new_univ, NaiveBayes))
	numpy.testing.assert_array_equal( model.weights, new_univ.weights)

@with_setup(setup_multivariate_gaussian, teardown)
def test_io_log_probability():
	X2 = DataGenerator(X)
	X3 = DataFrameGenerator(pandas.DataFrame(X))
	logp1 = model.log_probability(X)
	logp2 = model.log_probability(X2)
	logp3 = model.log_probability(X3)
	assert_array_almost_equal(logp1, logp2)
	assert_array_almost_equal(logp1, logp3)

@with_setup(setup_multivariate_gaussian, teardown)
def test_io_predict():
	X2 = DataGenerator(X)
	X3 = DataFrameGenerator(pandas.DataFrame(X))
	y_hat1 = model.predict(X)
	y_hat2 = model.predict(X2)
	y_hat3 = model.predict(X3)
	assert_array_almost_equal(y_hat1, y_hat2)
	assert_array_almost_equal(y_hat1, y_hat3)

@with_setup(setup_multivariate_gaussian, teardown)
def test_io_predict_proba():
	X2 = DataGenerator(X)
	X3 = DataFrameGenerator(pandas.DataFrame(X))
	y_hat1 = model.predict_proba(X)
	y_hat2 = model.predict_proba(X2)
	y_hat3 = model.predict_proba(X3)
	assert_array_almost_equal(y_hat1, y_hat2)
	assert_array_almost_equal(y_hat1, y_hat3)

@with_setup(setup_multivariate_gaussian, teardown)
def test_io_predict_log_proba():
	X2 = DataGenerator(X)
	X3 = DataFrameGenerator(pandas.DataFrame(X))
	y_hat1 = model.predict_log_proba(X)
	y_hat2 = model.predict_log_proba(X2)
	y_hat3 = model.predict_log_proba(X3)
	assert_array_almost_equal(y_hat1, y_hat2)
	assert_array_almost_equal(y_hat1, y_hat3)

def test_io_fit():
	X = numpy.random.randn(100, 5) + 0.5
	weights = numpy.abs(numpy.random.randn(100))
	y = numpy.random.randint(2, size=100)
	data_generator = DataGenerator(X, weights, y)

	d1 = IndependentComponentsDistribution([
		NormalDistribution(0, 1) for i in range(5)])
	d2 = IndependentComponentsDistribution([
		NormalDistribution(1, 1) for i in range(5)])

	nb1 = NaiveBayes([d1, d2])
	nb1.fit(X, y, weights)

	d1 = IndependentComponentsDistribution([
		NormalDistribution(0, 1) for i in range(5)])
	d2 = IndependentComponentsDistribution([
		NormalDistribution(1, 1) for i in range(5)])

	nb2 = NaiveBayes([d1, d2])
	nb2.fit(data_generator)

	logp1 = nb1.log_probability(X)
	logp2 = nb2.log_probability(X)

	assert_array_almost_equal(logp1, logp2)

def test_io_from_samples():
	X = numpy.random.randn(100, 5) + 0.5
	weights = numpy.abs(numpy.random.randn(100))
	y = numpy.random.randint(2, size=100)
	data_generator = DataGenerator(X, weights, y)

	d = NormalDistribution

	nb1 = NaiveBayes.from_samples(d, X=X, weights=weights, y=y)
	nb2 = NaiveBayes.from_samples(d, X=data_generator)

	logp1 = nb1.log_probability(X)
	logp2 = nb2.log_probability(X)

	assert_array_almost_equal(logp1, logp2)


from __future__ import  (division, print_function)

from pomegranate import *
from nose.tools import with_setup
from nose.tools import assert_equal
from nose.tools import assert_not_equal
from nose.tools import assert_raises
from nose.tools import assert_almost_equal
import random
import numpy as np
import json

def setup():
	'''
	Build a model that we want to use to test sequences. This model will
	be somewhat complicated, in order to extensively test YAHMM. This will be
	a three state global sequence alignment HMM. The HMM models a reference of
	'ACT', with pseudocounts to allow for slight deviations from this
	reference.
	'''

	random.seed(0)

	global model
	model = HiddenMarkovModel( "Global Alignment")

	# Define the distribution for insertions
	i_d = DiscreteDistribution( { 'A': 0.25, 'C': 0.25, 'G': 0.25, 'T': 0.25 } )

	# Create the insert states
	i0 = State( i_d, name="I0" )
	i1 = State( i_d, name="I1" )
	i2 = State( i_d, name="I2" )
	i3 = State( i_d, name="I3" )

	# Create the match states
	m1 = State( DiscreteDistribution({ "A": 0.95, 'C': 0.01, 'G': 0.01, 'T': 0.02 }) , name="M1" )
	m2 = State( DiscreteDistribution({ "A": 0.003, 'C': 0.99, 'G': 0.003, 'T': 0.004 }) , name="M2" )
	m3 = State( DiscreteDistribution({ "A": 0.01, 'C': 0.01, 'G': 0.01, 'T': 0.97 }) , name="M3" )

	# Create the delete states
	d1 = State( None, name="D1" )
	d2 = State( None, name="D2" )
	d3 = State( None, name="D3" )

	# Add all the states to the model
	model.add_states( [i0, i1, i2, i3, m1, m2, m3, d1, d2, d3 ] )

	# Create transitions from match states
	model.add_transition( model.start, m1, 0.9 )
	model.add_transition( model.start, i0, 0.1 )
	model.add_transition( m1, m2, 0.9 )
	model.add_transition( m1, i1, 0.05 )
	model.add_transition( m1, d2, 0.05 )
	model.add_transition( m2, m3, 0.9 )
	model.add_transition( m2, i2, 0.05 )
	model.add_transition( m2, d3, 0.05 )
	model.add_transition( m3, model.end, 0.9 )
	model.add_transition( m3, i3, 0.1 )

	# Create transitions from insert states
	model.add_transition( i0, i0, 0.70 )
	model.add_transition( i0, d1, 0.15 )
	model.add_transition( i0, m1, 0.15 )

	model.add_transition( i1, i1, 0.70 )
	model.add_transition( i1, d2, 0.15 )
	model.add_transition( i1, m2, 0.15 )

	model.add_transition( i2, i2, 0.70 )
	model.add_transition( i2, d3, 0.15 )
	model.add_transition( i2, m3, 0.15 )

	model.add_transition( i3, i3, 0.85 )
	model.add_transition( i3, model.end, 0.15 )

	# Create transitions from delete states
	model.add_transition( d1, d2, 0.15 )
	model.add_transition( d1, i1, 0.15 )
	model.add_transition( d1, m2, 0.70 )

	model.add_transition( d2, d3, 0.15 )
	model.add_transition( d2, i2, 0.15 )
	model.add_transition( d2, m3, 0.70 )

	model.add_transition( d3, i3, 0.30 )
	model.add_transition( d3, model.end, 0.70 )

	# Call bake to finalize the structure of the model.
	model.bake()


def multitransition_setup():
	'''
	Build a model that we want to use to test sequences. This is the same as the
	above model, except that it uses the multiple transition methods for building.
	'''

	random.seed(0)

	global model
	model = HiddenMarkovModel( "Global Alignment")

	# Define the distribution for insertions
	i_d = DiscreteDistribution( { 'A': 0.25, 'C': 0.25, 'G': 0.25, 'T': 0.25 } )

	# Create the insert states
	i0 = State( i_d, name="I0" )
	i1 = State( i_d, name="I1" )
	i2 = State( i_d, name="I2" )
	i3 = State( i_d, name="I3" )

	# Create the match states
	m1 = State( DiscreteDistribution({ "A": 0.95, 'C': 0.01, 'G': 0.01, 'T': 0.02 }) , name="M1" )
	m2 = State( DiscreteDistribution({ "A": 0.003, 'C': 0.99, 'G': 0.003, 'T': 0.004 }) , name="M2" )
	m3 = State( DiscreteDistribution({ "A": 0.01, 'C': 0.01, 'G': 0.01, 'T': 0.97 }) , name="M3" )

	# Create the delete states
	d1 = State( None, name="D1" )
	d2 = State( None, name="D2" )
	d3 = State( None, name="D3" )

	# Add all the states to the model
	model.add_states( [i0, i1, i2, i3, m1, m2, m3, d1, d2, d3 ] )

	# Create transitions from match states
	model.add_transitions( model.start, [m1, i0], [0.9, 0.1] )

	model.add_transitions( m1, [m2, i1, d2], [0.9, 0.05, 0.05] )
	model.add_transitions( m2, [m3, i2, d3], [0.9, 0.05, 0.05] )
	model.add_transitions( m3, [model.end, i3], [0.9, 0.1] )

	# Create transitions from insert states
	model.add_transitions( i0, [i0, d1, m1], [0.7, 0.15, 0.15] )
	model.add_transitions( i1, [i1, d2, m2], [0.7, 0.15, 0.15] )
	model.add_transitions( i2, [i2, d3, m3], [0.7, 0.15, 0.15] )
	model.add_transitions( [i3, i3], [i3, model.end], [0.85, 0.15] )

	# Create transitions from delete states
	model.add_transitions( d1, [d2, i1, m2], [0.15, 0.15, 0.70] )
	model.add_transitions( [d2, d2, d2, d3, d3], [d3, i2, m3, i3, model.end],
		[0.15, 0.15, 0.70, 0.30, 0.70 ] )

	# Call bake to finalize the structure of the model.
	model.bake()


def tied_edge_setup():
	'''
	Build a model that we want to use to test sequences. This model has
	tied edges.
	'''

	random.seed(0)

	global model
	model = HiddenMarkovModel( "Global Alignment")

	# Define the distribution for insertions
	i_d = DiscreteDistribution( { 'A': 0.25, 'C': 0.25, 'G': 0.25, 'T': 0.25 } )

	# Create the insert states
	i0 = State( i_d, name="I0" )
	i1 = State( i_d, name="I1" )
	i2 = State( i_d, name="I2" )
	i3 = State( i_d, name="I3" )

	# Create the match states
	m1 = State( DiscreteDistribution({ "A": 0.95, 'C': 0.01, 'G': 0.01, 'T': 0.02 }) , name="M1" )
	m2 = State( DiscreteDistribution({ "A": 0.003, 'C': 0.99, 'G': 0.003, 'T': 0.004 }) , name="M2" )
	m3 = State( DiscreteDistribution({ "A": 0.01, 'C': 0.01, 'G': 0.01, 'T': 0.97 }) , name="M3" )

	# Create the delete states
	d1 = State( None, name="D1" )
	d2 = State( None, name="D2" )
	d3 = State( None, name="D3" )

	# Add all the states to the model
	model.add_states( [i0, i1, i2, i3, m1, m2, m3, d1, d2, d3 ] )

	# Create transitions from match states
	model.add_transition( model.start, m1, 0.9 )
	model.add_transition( model.start, i0, 0.1 )
	model.add_transition( m1, m2, 0.9 )
	model.add_transition( m1, i1, 0.05 )
	model.add_transition( m1, d2, 0.05 )
	model.add_transition( m2, m3, 0.9 )
	model.add_transition( m2, i2, 0.05 )
	model.add_transition( m2, d3, 0.05 )
	model.add_transition( m3, model.end, 0.9 )
	model.add_transition( m3, i3, 0.1 )

	# Create transitions from insert states
	model.add_transition( i0, i0, 0.70, group="i_a" )
	model.add_transition( i0, d1, 0.15, group="i_b" )
	model.add_transition( i0, m1, 0.15, group="i_c" )

	model.add_transition( i1, i1, 0.70, group="i_a" )
	model.add_transition( i1, d2, 0.15, group="i_b" )
	model.add_transition( i1, m2, 0.15, group="i_c" )

	model.add_transition( i2, i2, 0.70, group="i_a" )
	model.add_transition( i2, d3, 0.15, group="i_b" )
	model.add_transition( i2, m3, 0.15, group="i_c" )

	model.add_transition( i3, i3, 0.85, group="i_a" )
	model.add_transition( i3, model.end, 0.15 )

	# Create transitions from delete states
	model.add_transition( d1, d2, 0.15, group="d_a" )
	model.add_transition( d1, i1, 0.15, group="d_b" )
	model.add_transition( d1, m2, 0.70, group="d_c" )

	model.add_transition( d2, d3, 0.15, group="d_a" )
	model.add_transition( d2, i2, 0.15, group="d_b" )
	model.add_transition( d2, m3, 0.70, group="d_c" )

	model.add_transition( d3, i3, 0.30 )
	model.add_transition( d3, model.end, 0.70 )

	# Call bake to finalize the structure of the model.
	model.bake()


def teardown():
	'''
	Remove the model at the end of the unit testing. Since it is stored in a
	global variance, simply delete it.
	'''

	pass


@with_setup( setup, teardown )
def test_same_length_viterbi():
	scores = [ -0.5132449003570658, -11.048101241343396, -9.125519674022627,
		-5.0879558788604475 ]
	sequences = [ list(x) for x in [ 'ACT', 'GGC', 'GAT', 'ACC' ] ]

	for seq, score in zip( sequences, scores ):
		assert_almost_equal( model.viterbi( seq )[0], score )

	assert_raises( ValueError, model.viterbi, list('XXX') )


@with_setup( setup, teardown )
def test_variable_length_viterbi():
	scores = [ -5.406181012423981, -10.88681993576597, -3.6244718790494277,
	-3.644880750680635, -10.674332964640293, -10.393824835172445,
	-8.67126440174503, -16.903451796110275, -16.451699654050792 ]
	sequences = [ list(x) for x in ('A', 'GA', 'AC', 'AT', 'ATCC',
		'ACGTG', 'ATTT', 'TACCCTC', 'TGTCAACACT') ]

	for seq, score in zip( sequences, scores ):
		assert_almost_equal( model.viterbi( seq )[0], score )


@with_setup( setup, teardown )
def test_log_probability():
	scores = [ -5.3931, -0.5052, -11.8478, -14.3482 ]
	sequences = [ list(x) for x in ( 'A', 'ACT', 'GGCA', 'TACCTGT' ) ]

	for seq, score in zip( sequences, scores ):
		assert_equal( round( model.log_probability( seq ), 4 ), score )


@with_setup( setup, teardown )
def test_posterior_transitions():
	a_scores = [ 0.0, 0.0021, 0.2017, 1.5105 ]
	b_scores = [ 0.013, 0.0036, 1.9836, 2.145 ]
	c_scores = [ 0.013, 0.0035, 0.817, 0.477 ]
	d_scores = [ 1.0, 0.0023, 0.2636, 0.3682 ]
	t_scores = [ 4.013, 4.0083, 6.457, 8.9812 ]
	sequences = [ list(x) for x in ( 'A', 'ACT', 'GGCA', 'TACCTGT' ) ]

	indices = { state.name: i for i, state in enumerate( model.states ) }
	i, j, k, l = indices['I2'], indices['I0'], indices['D1'], indices['D2']

	scores = zip( sequences, a_scores, b_scores, c_scores, d_scores, t_scores )
	for seq, a, b, c, d, t in scores:
		trans, ems = model.forward_backward( seq )

		assert_equal( round( trans[i].sum(), 4 ), a )
		assert_equal( round( trans[j].sum(), 4 ), b )
		assert_equal( round( trans[k].sum(), 4 ), c )
		assert_equal( round( trans[l].sum(), 4 ), d )
		assert_equal( round( trans.sum(), 4 ), t )


@with_setup( setup, teardown )
def test_posterior_transitions_w_training():
	sequences = [ list(x) for x in ( 'A', 'ACT', 'GGCA', 'TACCTGT' ) ]
	indices = { state.name: i for i, state in enumerate( model.states ) }

	transitions = model.dense_transition_matrix()
	i0, i1, i2 = indices['I0'], indices['I1'], indices['I2']
	d1, d2, d3 = indices['D1'], indices['D2'], indices['D3']
	m1, m2, m3 = indices['M1'], indices['M2'], indices['M3']

	assert_equal( transitions[d1, i1], transitions[d2, i2] )
	assert_equal( transitions[i0, i0], transitions[i1, i1] )
	assert_equal( transitions[i0, i0], transitions[i2, i2] )
	assert_equal( transitions[i0, m1], transitions[i1, m2] )
	assert_equal( transitions[d1, d2], transitions[d2, d3] )
	assert_equal( transitions[i0, d1], transitions[i1, d2] )
	assert_equal( transitions[i0, d1], transitions[i2, d3] )

	model.fit( sequences, verbose=False )
	transitions = model.dense_transition_matrix()

	assert_not_equal( transitions[d1, i1], transitions[d2, i2] )
	assert_not_equal( transitions[i0, m1], transitions[i1, m2] )
	assert_not_equal( transitions[d1, d2], transitions[d2, d3] )
	assert_not_equal( transitions[i0, d1], transitions[i1, d2] )
	assert_not_equal( transitions[i0, d1], transitions[i2, d3] )


@with_setup( setup, teardown )
def test_posterior_transitions_w_vtraining():
	sequences = [ list(x) for x in ( 'A', 'ACT', 'GGCA', 'TACCTGT' ) ]
	indices = { state.name: i for i, state in enumerate( model.states ) }

	transitions = model.dense_transition_matrix()
	i0, i1, i2, i3 = indices['I0'], indices['I1'], indices['I2'], indices['I3']
	d1, d2, d3 = indices['D1'], indices['D2'], indices['D3']
	m1, m2, m3 = indices['M1'], indices['M2'], indices['M3']

	assert_equal( transitions[d1, i1], transitions[d2, i2] )
	assert_equal( transitions[i0, i0], transitions[i1, i1] )
	assert_equal( transitions[i0, i0], transitions[i2, i2] )
	assert_equal( transitions[i0, m1], transitions[i1, m2] )
	assert_equal( transitions[d1, d2], transitions[d2, d3] )
	assert_equal( transitions[i0, d1], transitions[i1, d2] )
	assert_equal( transitions[i0, d1], transitions[i2, d3] )

	model.fit( sequences, verbose=False, algorithm='viterbi' )
	transitions = model.dense_transition_matrix()

	assert_not_equal( transitions[i0, i0], transitions[i1, i1] )
	assert_not_equal( transitions[d1, d2], transitions[d2, d3] )
	assert_not_equal( transitions[i0, d1], transitions[i1, d2] )
	assert_not_equal( transitions[i0, d1], transitions[i2, d3] )


@with_setup( tied_edge_setup, teardown )
def test_posterior_transitions_w_tied_training():
	sequences = [ list(x) for x in ( 'A', 'ACT', 'GGCA', 'TACCTGT' ) ]
	indices = { state.name: i for i, state in enumerate( model.states ) }

	transitions = model.dense_transition_matrix()
	i0, i1, i2, i3 = indices['I0'], indices['I1'], indices['I2'], indices['I3']
	d1, d2, d3 = indices['D1'], indices['D2'], indices['D3']
	m1, m2, m3 = indices['M1'], indices['M2'], indices['M3']

	assert_equal( transitions[d1, i1], transitions[d2, i2] )
	assert_equal( transitions[i0, i0], transitions[i1, i1] )
	assert_equal( transitions[i0, i0], transitions[i2, i2] )
	assert_equal( transitions[i0, m1], transitions[i1, m2] )
	assert_equal( transitions[d1, d2], transitions[d2, d3] )
	assert_equal( transitions[i0, d1], transitions[i1, d2] )
	assert_equal( transitions[i0, d1], transitions[i2, d3] )

	model.fit( sequences, verbose=False )
	transitions = model.dense_transition_matrix()

	assert_equal( transitions[i0, i0], transitions[i1, i1] )
	assert_equal( transitions[d1, d2], transitions[d2, d3] )
	assert_equal( transitions[i0, d1], transitions[i1, d2] )
	assert_equal( transitions[i0, d1], transitions[i2, d3] )


@with_setup( tied_edge_setup, teardown )
def test_posterior_transitions_w_tied_vtraining():
	sequences = [ list(x) for x in ( 'A', 'ACT', 'GGCA', 'TACCTGT' ) ]
	indices = { state.name: i for i, state in enumerate( model.states ) }

	transitions = model.dense_transition_matrix()
	i0, i1, i2 = indices['I0'], indices['I1'], indices['I2']
	d1, d2, d3 = indices['D1'], indices['D2'], indices['D3']
	m1, m2, m3 = indices['M1'], indices['M2'], indices['M3']

	assert_equal( transitions[d1, i1], transitions[d2, i2] )
	assert_equal( transitions[i0, i0], transitions[i1, i1] )
	assert_equal( transitions[i0, i0], transitions[i2, i2] )
	assert_equal( transitions[i0, m1], transitions[i1, m2] )
	assert_equal( transitions[d1, d2], transitions[d2, d3] )
	assert_equal( transitions[i0, d1], transitions[i1, d2] )
	assert_equal( transitions[i0, d1], transitions[i2, d3] )

	model.fit( sequences, verbose=False, algorithm='viterbi' )
	transitions = model.dense_transition_matrix()

	assert_equal( transitions[d1, i1], transitions[d2, i2] )
	assert_equal( transitions[i0, i0], transitions[i1, i1] )
	assert_equal( transitions[i0, i0], transitions[i2, i2] )
	assert_equal( transitions[i0, m1], transitions[i1, m2] )
	assert_equal( transitions[d1, d2], transitions[d2, d3] )
	assert_equal( transitions[i0, d1], transitions[i1, d2] )
	assert_equal( transitions[i0, d1], transitions[i2, d3] )


@with_setup( setup, teardown )
def test_posterior_emissions():
	a_scores = [ 0.987, 0.9965, 0.183, 0.523 ]
	b_scores = [ 0.0, 0.9977, 0.7364, 0.6318 ]
	c_scores = [ 0.0, 0.9975, 0.6237, 0.8641 ]
	d_scores = [ 0.0, 0.0021, 0.2017, 1.5105 ]
	sequences = [ list(x) for x in ( 'A', 'ACT', 'GGCA', 'TACCTGT' ) ]

	indices = { state.name: i for i, state in enumerate( model.states ) }
	i, j, k, l = indices['M1'], indices['M2'], indices['M3'], indices['I2']

	for seq, a, b, c, d in zip( sequences, a_scores, b_scores, c_scores, d_scores ):
		trans, ems = model.forward_backward( seq )
		ems = np.exp( ems )

		assert_equal( round( ems[:,i].sum(), 4 ), a )
		assert_equal( round( ems[:,j].sum(), 4 ), b )
		assert_equal( round( ems[:,k].sum(), 4 ), c )
		assert_equal( round( ems[:,l].sum(), 4 ), d )
		assert_equal( round( ems.sum() ), len( seq ) )


@with_setup( multitransition_setup, teardown )
def test_posterior_emissions_w_multitransition_setup():
	a_scores = [ 0.987, 0.9965, 0.183, 0.523 ]
	b_scores = [ 0.0, 0.9977, 0.7364, 0.6318 ]
	c_scores = [ 0.0, 0.9975, 0.6237, 0.8641 ]
	d_scores = [ 0.0, 0.0021, 0.2017, 1.5105 ]
	sequences = [ list(x) for x in ( 'A', 'ACT', 'GGCA', 'TACCTGT' ) ]

	indices = { state.name: i for i, state in enumerate( model.states ) }
	i, j, k, l = indices['M1'], indices['M2'], indices['M3'], indices['I2']

	for seq, a, b, c, d in zip( sequences, a_scores, b_scores, c_scores, d_scores ):
		trans, ems = model.forward_backward( seq )
		ems = np.exp( ems )

		assert_equal( round( ems[:,i].sum(), 4 ), a )
		assert_equal( round( ems[:,j].sum(), 4 ), b )
		assert_equal( round( ems[:,k].sum(), 4 ), c )
		assert_equal( round( ems[:,l].sum(), 4 ), d )
		assert_equal( round( ems.sum() ), len( seq ) )


@with_setup( tied_edge_setup, teardown )
def test_posterior_emissions_w_tied_edge_setup():
	a_scores = [ 0.987, 0.9965, 0.183, 0.523 ]
	b_scores = [ 0.0, 0.9977, 0.7364, 0.6318 ]
	c_scores = [ 0.0, 0.9975, 0.6237, 0.8641 ]
	d_scores = [ 0.0, 0.0021, 0.2017, 1.5105 ]
	sequences = [ list(x) for x in ( 'A', 'ACT', 'GGCA', 'TACCTGT' ) ]

	indices = { state.name: i for i, state in enumerate( model.states ) }
	i, j, k, l = indices['M1'], indices['M2'], indices['M3'], indices['I2']

	for seq, a, b, c, d in zip( sequences, a_scores, b_scores, c_scores, d_scores ):
		trans, ems = model.forward_backward( seq )
		ems = np.exp( ems )

		assert_equal( round( ems[:,i].sum(), 4 ), a )
		assert_equal( round( ems[:,j].sum(), 4 ), b )
		assert_equal( round( ems[:,k].sum(), 4 ), c )
		assert_equal( round( ems[:,l].sum(), 4 ), d )
		assert_equal( round( ems.sum() ), len( seq ) )


@with_setup( setup, teardown )
def test_properties():
	assert_equal( model.edge_count(), 29 )
	assert_equal( model.state_count(), 12 )
	assert_equal( model.name, "Global Alignment" )


@with_setup( setup, teardown )
def test_to_json():
	b = json.loads(model.to_json())

	assert_equal(b['name'], 'Global Alignment')
	assert_equal(len(b['edges']), 29)
	assert_equal(len(b['states']), 12)
	assert_equal(b['silent_index'], 7)


@with_setup( setup, teardown )
def test_from_json():
	hmm = HiddenMarkovModel.from_json( model.to_json() )

	assert_equal(hmm.edge_count(), 29)
	assert_equal(hmm.state_count(), 12)
	assert_equal(hmm.name, "Global Alignment")



#!python
#cython: boundscheck=False
#cython: cdivision=True
# JointProbabilityTable.pyx
# Contact: Jacob Schreiber <jmschreiber91@gmail.com>

from libc.stdlib cimport calloc
from libc.stdlib cimport free
from libc.stdlib cimport malloc
from libc.string cimport memset
from libc.math cimport exp as cexp

from ..utils cimport _log
from ..utils cimport isnan
from ..utils import check_random_state
from ..utils import _check_nan

import itertools as it
import json
import numpy
import random
import scipy

from .DiscreteDistribution import DiscreteDistribution

cdef class JointProbabilityTable(MultivariateDistribution):
	"""
	A joint probability table. The primary difference between this and the
	conditional table is that the final column sums to one here. The joint
	table can be thought of as the conditional probability table normalized
	by the marginals of each parent.
	"""

	def __cinit__(self, table, parents=None, frozen=False):
		"""
		Take in the distribution represented as a list of lists, where each
		inner list represents a row.
		"""

		self.name = "JointProbabilityTable"
		self.frozen = False
		self.d = len(parents) if parents is not None else len(table[0]) - 1
		self.m = len(parents) if parents is not None else len(table[0]) - 1
		self.n = len(table)
		self.k = len(set(row[-2] for row in table))
		self.idxs = <int*> malloc((self.m+1)*sizeof(int))

		self.values = <double*> malloc(self.n*sizeof(double))
		self.counts = <double*> calloc(self.n, sizeof(double))
		self.count = 0

		self.n_columns = self.d

		self.dtypes = []
		for column in table[0]:
			dtype = str(type(column)).split()[-1].strip('>').strip("'")
			self.dtypes.append(dtype)

		self.idxs[0] = 1
		for i in range(self.m-1):
			self.idxs[i+1] = len(set(row[self.m-i-2] for row in table))
		self.idxs[self.m] = 0

		self.keymap = {}
		for i, row in enumerate(table):
			self.keymap[tuple(row[:-1])] = i
			self.values[i] = _log(row[-1])

		self.parents = list(parents) if parents is not None else None
		self.parameters = [[list(row) for row in table], self.parents, self.keymap]

	def __dealloc__(self):
		free(self.idxs)
		free(self.values)
		free(self.counts)

	def __reduce__(self):
		return self.__class__, (self.parameters[0], self.parents, self.frozen)

	def __str__(self):
		return "\n".join(
					"\t".join(map(str, key + (cexp(self.values[idx]),)))
							for key, idx in self.keymap.items())

	def __len__(self):
		return self.k

	def sample(self, n=None, random_state=None):
		random_state = check_random_state(random_state)
		a = random_state.uniform(0, 1)
		values = numpy.cumsum(numpy.exp([self.values[i] for i in range(self.n)]))
		for i in range(self.n):
			if values[i] > a:
				return list(self.keymap.keys())[i]

	def bake(self, keys):
		"""Order the inputs according to some external global ordering."""

		keymap, values = [], []
		for i, key in enumerate(keys):
			keymap.append((key, i))
			idx = self.keymap[key]
			values.append(self.values[idx])

		for i in range(len(keys)):
			self.values[i] = values[i]

		self.keymap = dict(keymap)

	def keys(self):
		return tuple(row for row in self.parameters[2].keys())

	def log_probability(self, X):
		"""
		Return the log probability of a value, which is a tuple in proper
		ordering, like the training data.
		"""

		X = numpy.array(X, ndmin=2, dtype=object)

		log_probabilities = numpy.zeros(X.shape[0])
		for i, x in enumerate(X):
			x = tuple(x)

			for x_ in x:
				if _check_nan(x_):
					break
			else:
				key = self.keymap[x]
				log_probabilities[i] = self.values[key]

		if X.shape[0] == 1:
			return log_probabilities[0]
		return log_probabilities

	cdef void _log_probability(self, double* X, double* log_probability, 
		int n) nogil:
		cdef int i, j, idx

		for i in range(n):
			idx = 0
			for j in range(self.m+1):
				if isnan(X[self.m-j]):
					log_probability[i] = 0.
					break

				idx += self.idxs[j] * <int> X[self.m-j]
			else:
				log_probability[i] = self.values[idx]

	def marginal(self, wrt=-1, neighbor_values=None):
		"""
		Determine the marginal of this table with respect to the index of one
		variable. The parents are index 0..n-1 for n parents, and the final
		variable is either the appropriate value or -1. For example:
		table =
		A    B    C    p(C)
		... data ...
		table.marginal(0) gives the marginal wrt A
		table.marginal(1) gives the marginal wrt B
		table.marginal(2) gives the marginal wrt C
		table.marginal(-1) gives the marginal wrt C
		"""

		if isinstance(neighbor_values, dict):
			neighbor_values = [neighbor_values.get(d, None) for d in self.parents]

		if isinstance(neighbor_values, list):
			wrt = neighbor_values.index(None)

		# Determine the keys for the respective parent distribution
		d = {k: 0 for k in self.parents[wrt].keys()}
		total = 0.0

		for key, idx in self.keymap.items():
			logp = self.values[idx]

			if neighbor_values is not None:
				for j, k in enumerate(key):
					if j == wrt:
						continue

					logp += neighbor_values[j].log_probability(k)

			p = cexp(logp)
			d[key[wrt]] += p
			total += p

		for key, value in d.items():
			d[key] = value / total if total > 0 else 1. / len(self.parents[wrt].keys())

		return DiscreteDistribution(d)

	def summarize(self, items, weights=None):
		"""Summarize the data into sufficient statistics to store."""

		if len(items) == 0 or self.frozen == True:
			return

		if weights is None:
			weights = numpy.ones(len(items), dtype='float64')
		elif numpy.sum(weights) == 0:
			return
		else:
			weights = numpy.asarray(weights, dtype='float64')

		self.__summarize(items, weights)

	cdef void __summarize(self, items, double [:] weights):
		cdef int i, n = len(items)
		cdef tuple item

		for i in range(n):
			item = tuple(items[i])

			if _check_nan(item):
				continue

			key = self.keymap[item]
			self.counts[key] += weights[i]
			self.count += weights[i]

	cdef double _summarize(self, double* items, double* weights, int n,
		int column_idx, int d) nogil:
		cdef int i, j, idx
		cdef double count = 0
		cdef double* counts = <double*> calloc(self.n, sizeof(double))

		for i in range(n):
			idx = 0
			for j in range(self.m+1):
				if isnan(items[self.m-i]):
					break

				idx += self.idxs[i] * <int> items[self.m-i]
			else:
				counts[idx] += weights[i]
				count += weights[i]

		with gil:
			self.count += count
			for i in range(self.n):
				self.counts[i] += counts[i]

		free(counts)

	def from_summaries(self, inertia=0.0, pseudocount=0.0):
		"""Update the parameters of the table."""

		cdef int i, k

		w_sum = sum(self.counts[i] for i in range(self.n))
		if w_sum < 1e-7:
			return

		for i in range(self.n):
			probability = ((self.counts[i] + pseudocount) / 
				(self.count + pseudocount * self.k))
			
			self.values[i] = _log(cexp(self.values[i])*inertia +
				probability*(1-inertia))

		for i in range(self.n):
			self.parameters[0][i][-1] = cexp(self.values[i])

		self.clear_summaries()

	def clear_summaries(self):
		"""Clear the summary statistics stored in the object."""

		self.count = 0
		memset(self.counts, 0, self.n*sizeof(double))

	def fit(self, items, weights=None, inertia=0.0, pseudocount=0.0):
		"""Update the parameters of the table based on the data."""

		self.summarize(items, weights)
		self.from_summaries(inertia, pseudocount)

	def to_json(self, separators=(',', ' : '), indent=4):
		"""Serialize the model to a JSON.

		Parameters
		----------
		separators : tuple, optional
		    The two separators to pass to the json.dumps function for formatting.
		    Default is (',', ' : ').

		indent : int, optional
		    The indentation to use at each level. Passed to json.dumps for
		    formatting. Default is 4.

		Returns
		-------
		json : str
		    A properly formatted JSON object.
		"""

		table = [list(key + tuple([cexp(self.values[i])])) for key, i in self.keymap.items()]

		model = {
					'class' : 'Distribution',
		            'name' : 'JointProbabilityTable',
		            'table' : table,
		            'dtypes' : self.dtypes,
		            'parents' : [dist if isinstance(dist, int) else json.loads(dist.to_json()) for dist in self.parameters[1]]
		        }
		
		return json.dumps(model, separators=separators, indent=indent)

	@classmethod
	def from_samples(cls, X, parents=None, weights=None, pseudocount=0.0, 
		keys=None):
		"""Learn the table from data."""

		X = numpy.array(X)
		n, d = X.shape

		if parents is None:
			parents = list(range(X.shape[1]))

		keys = keys or [numpy.unique(X[:,i]).tolist() for i in range(d)]
		m = numpy.prod([len(k) for k in keys])

		table = []
		for key in it.product(*keys):
			table.append(list(key) + [1./m,])

		d = cls(table, parents)
		d.fit(X, weights, pseudocount=pseudocount)
		return d



# DiscreteDistribution.pxd
# Contact: Jacob Schreiber <jmschreiber91@gmail.com>

import numpy
cimport numpy

from .distributions cimport Distribution

cdef class DiscreteDistribution(Distribution):
	cdef bint encoded_summary
	cdef int n
	cdef str dtype
	cdef dict dist, log_dist
	cdef tuple encoded_keys
	cdef double* encoded_counts
	cdef double* encoded_log_probability
	cdef double __probability(self, symbol)
	cdef double __log_probability(self, symbol)




from setuptools import setup
from setuptools import Extension
from setuptools.command.build_ext import build_ext as _build_ext
# https://stackoverflow.com/a/11181607/541202
# import __builtin__ as __builtins__

try:
    from Cython.Build import cythonize
except ImportError:
    use_cython = False
    ext = 'c'
else:
    use_cython = True
    ext = 'pyx'

filenames = [ 
    "base",
    "bayes",
    "BayesianNetwork",
    "MarkovNetwork",
    "FactorGraph",
    "hmm",
    "gmm",
    "kmeans",
    "NaiveBayes",
    "BayesClassifier",
    "MarkovChain",
    "utils",
    "parallel"
]

distributions = [
    'distributions',
    'UniformDistribution',
    'BernoulliDistribution',
    'NormalDistribution',
    'LogNormalDistribution',
    'ExponentialDistribution',
    'BetaDistribution',
    'GammaDistribution',
    'DiscreteDistribution',
    'PoissonDistribution',
    'KernelDensities',
    'IndependentComponentsDistribution',
    'MultivariateGaussianDistribution',
    'DirichletDistribution',
    'ConditionalProbabilityTable',
    'JointProbabilityTable',
    "NeuralNetworkWrapper"
]

if not use_cython:
    extensions = [
        Extension("pomegranate.{}".format( name ), [ "pomegranate/{}.{}".format(name, ext) ]) for name in filenames
    ] + [Extension("pomegranate.distributions.{}".format(dist), ["pomegranate/distributions/{}.{}".format(dist, ext)]) for dist in distributions]
else:
    extensions = [
            Extension("pomegranate.*", ["pomegranate/*.pyx"]),
	        Extension("pomegranate.distributions.*", ["pomegranate/distributions/*.pyx"])
    ]

    extensions = cythonize(extensions, compiler_directives={'language_level' : "2"})

class build_ext(_build_ext):
    def finalize_options(self):
        _build_ext.finalize_options(self)
        # Prevent numpy from thinking it is still in its setup process:
        __builtins__.__NUMPY_SETUP__ = False
        import numpy
        self.include_dirs.append(numpy.get_include())

setup(
    name='pomegranate',
    version='0.13.4',
    author='Jacob Schreiber',
    author_email='jmschreiber91@gmail.com',
    packages=[
        'pomegranate',
        'pomegranate/distributions',
    ],
    url='http://pypi.python.org/pypi/pomegranate/',
    license='LICENSE.txt',
    description='Pomegranate is a graphical models library for Python, implemented in Cython for speed.',
    ext_modules=extensions,
    cmdclass={'build_ext':build_ext},
    setup_requires=[
        "cython >= 0.22.1",
        "numpy >= 1.8.0",
        "scipy >= 0.17.0"
    ],
    install_requires=[
        "numpy >= 1.8.0",
        "joblib >= 0.9.0b4",
        "networkx >= 2.0",
        "scipy >= 0.17.0",
        "pyyaml"
    ],
    test_suite = 'nose.collector',
    package_data={
        'pomegranate': ['*.pyd', '*.pxd'],
        'pomegranate/distributions': ['*.pyd', '*.pxd'],
    },
    include_package_data=True,
    zip_safe=False,
)



#!python
#cython: boundscheck=False
#cython: cdivision=True
# BetaDistribution.pyx
# Contact: Jacob Schreiber <jmschreiber91@gmail.com>

import numpy

from ..utils cimport _log
from ..utils cimport isnan
from ..utils cimport lgamma
from ..utils import check_random_state

# Define some useful constants
DEF NEGINF = float("-inf")
DEF INF = float("inf")

cdef class BetaDistribution(Distribution):
	"""A beta-bernoulli distribution.

	This object is a beta-bernoulli distribution. This means that it uses
	a beta distribution to model the distribution of values that the rate
	value can take rather than it being a single number.

	This should not be confused with a Beta distribution by itself.
	"""

	property parameters:
		def __get__(self):
			return [self.alpha, self.beta]
		def __set__(self, parameters):
			alpha, beta = parameters
			self.alpha, self.beta = alpha, beta
			self.beta_norm = lgamma(alpha+beta) - lgamma(alpha) - lgamma(beta)

	def __init__(self, alpha, beta, frozen=False):
		self.alpha = alpha
		self.beta = beta
		self.beta_norm = lgamma(alpha+beta) - lgamma(alpha) - lgamma(beta)
		self.summaries = [0, 0]
		self.name = "BetaDistribution"
		self.frozen = frozen

	def __reduce__(self):
		"""Serialize distribution for pickling."""
		return self.__class__, (self.alpha, self.beta, self.frozen)

	cdef void _log_probability(self, double* X, double* log_probability, int n) nogil:
		cdef int i
		cdef double alpha = self.alpha
		cdef double beta = self.beta
		cdef double beta_norm = self.beta_norm

		for i in range(n):
			if isnan(X[i]):
				log_probability[i] = 0.
			else:
				log_probability[i] = beta_norm + (alpha-1)*_log(X[i]) + \
					(beta-1)*_log(1-X[i])

	def sample(self, n=None, random_state=None):
		random_state = check_random_state(random_state)
		return random_state.beta(self.alpha, self.beta, n)

	cdef double _summarize(self, double* items, double* weights, int n,
		int column_idx, int d) nogil:
		"""Cython optimized function for summarizing some data."""

		cdef int i
		cdef double alpha = 0, beta = 0
		cdef double item

		for i in range(n):
			item = items[i*d + column_idx]
			if isnan(item):
				continue

			if item == 1:
				alpha += weights[i]
			else:
				beta += weights[i]

		with gil:
			self.summaries[0] += alpha
			self.summaries[1] += beta

	def from_summaries(self, inertia=0.0):
		"""Use the summaries in order to update the distribution."""

		if self.frozen == True:
			return

		alpha, beta = self.summaries

		self.alpha = self.alpha*inertia + alpha*(1-inertia)
		self.beta = self.beta*inertia + beta*(1-inertia)
		self.beta_norm = lgamma(self.alpha+self.beta) - lgamma(self.alpha) - lgamma(self.beta)

		self.summaries = [0, 0]

	def clear_summaries(self):
		"""Clear the summary statistics stored in the object."""

		self.summaries = [0, 0]

	@classmethod
	def blank(cls):
		return cls(0, 0)
